#ifndef __INCFILE__
#define __INCFILE__

typedef enum http_method_s {
	HTTP_M_UNKNOWN,
	HTTP_M_OPTIONS,
	HTTP_M_GET,
	HTTP_M_HEAD,
	HTTP_M_POST,
	HTTP_M_PUT,
	HTTP_M_DELETE,
	HTTP_M_TRACE,
	HTTP_M_MAX
} http_method_t;

typedef struct methods {
	const short number;
	const char * const name;
	int len;
} methods_t;


const char * http_protocol_strings[] = {
	"HTTP/0.9", /* The position is crucial & magic! do not insert, only add */
	"HTTP/1.0",
	"HTTP/1.1",
	NULL
};

const methods_t http_methods [HTTP_M_MAX] = {
	{HTTP_M_UNKNOWN,	"UNKNOWN",	7},
	{HTTP_M_OPTIONS,	"OPTIONS",	7},
	{HTTP_M_GET,		"GET",		3},
	{HTTP_M_HEAD,		"HEAD",		4},
	{HTTP_M_POST,		"POST",		4},
	{HTTP_M_PUT,		"PUT",		3},
	{HTTP_M_DELETE,		"DELETE",	6},
	{HTTP_M_TRACE,		"TRACE",	5}
};

typedef enum http_versions {
	HTTP_V_UNKNOWN,
	HTTP_V_0_9,
	HTTP_V_1_0,
	HTTP_V_1_1
} http_version_t;

typedef struct http_mime_headers {
	int	content_length;
	int	transfer_encoding;
	int	connection_close;
	char	*sep;	/* THIS_STRING_SEPARATES */
} http_mime_header_t;

typedef struct http_request_s
{
	const char *message;
	unsigned int message_len;
	unsigned int parsed_len;

	/* request method */
	http_method_t method;
	const char *method_str;
	unsigned int method_len;

	/* request URI */
	const char *uri_str;
	unsigned int uri_len;

	/* http version */
	http_version_t version;
	const char *version_str;
	unsigned int version_len;

	/* cookies */
	unsigned int cookies;
	const char *cookies_str;
	unsigned int cookies_len;

	/* MIME header */
	http_mime_header_t mime;
} http_request_t;

typedef struct http_response_s
{
	/* http verison */
	int version;

	/* response status code */
	int status_code;

	/* MIME header */
	http_mime_header_t mime;
} http_response_t;


typedef struct http_read_ctl_block_s
{
	struct socket	*sock;		/* socket that message read from */
	char*		buffer;		/* buffer to store read message */
	int		len;		/* size of buffer */
	char*		info;		/* point to the current information */
	int		offset;		/* offset of remaining bytes */
	int		remaining;	/* remaining bytes not return */
	int		flag;		/* read flag */
} http_read_ctl_block_t;

#endif
