#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ctype.h>

#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <asm/uaccess.h>

#include "tcp_vs.h"
#include "incfile.h"

#define CR 13
#define LF 10
#define SP ' '

#define PARSE_OK		0
#define PARSE_ERROR		1

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

/*****************************RULE MODULE***************************/


#define STICKY_ON_IP 		1
#define STICKY_ON_IP_PORT 	2
#define NON_STICKY			3



/* rule matching structure */
struct ip_vs_cb_rule_field
{
  char  name[512];
  char  value[256];
};


typedef struct server
{
  char *name;
  u_long addr;
}server;

typedef struct ip_sticky
{
  u_long source_ip;
  u_long dest_ip;
}ip_sticky;


/*
** i need to remove this junk from chandra's code of xml parsing
*/
int 
   len_cr=2,
   len_lf=2,
   len_sp=1,
   len_la=1,
   len_ra=1,
   len_fs=1,
   len_ex=1,
   max_char_len=1;

char
   *c_cr="\r",
   *c_lf="\n",
   *c_sp=" ",
   *c_la="<",
   *c_ra=">",
   *c_fs="/",
   *c_ex="!";


struct ip_vs_cb_rule_field rule_fields[] = {
						{"purchase:1.totalAmount:1.", "" },
						{"purchase:1.subTotal:1.", "" },
						{"purchase:1.subTotal:2.", "" },
						{"purchase:1.unitPrice:2.", "" }
					   };
#if 0
static struct server server_list[] =
{
  {"eca.uccs.edu",         0x80c6c0bc},  /*128.198.192.198*/
  {"frodo.uccs.edu",       0x80c6c0b7},  /*128.198.192.183*/
  {"wallace.uccs.edu",     0x80c6c0d0},  /*128.198.192.183*/
  {"reject",               0xffffffff}
};

int num_servers = sizeof(server_list) / sizeof(struct server);
#endif
int num_rule_fields = sizeof(rule_fields) / sizeof(struct ip_vs_cb_rule_field);



struct ip_sticky ip_sticky_list[20];
int sticky_num = 0;

char *mstrstr(char *source,  char *substr, size_t n);
int  get_content_type(char *buf, int buf_len, char *htp_mthd_nm);
int  xml_tag_parser(char *buf, int buf_len);
int  get_tag_range(char *tag, int pos, char *hay_begin, char *hay_end, char **needle_begin, char **needle_end);
void  clear_rule_fld_values(void);
char *strDecode(char *from);
u_long rule_configure(char *url, u_long saddr, u_short sport);
int atoi(char *p);

__u32 check_sticky(__u32 addr)
{
  int i = 0;

  for(i = 0; i < sticky_num; i++)
  {
    if(addr == ip_sticky_list[i].source_ip)
    {
      printk("sticky found on %X\n",ip_sticky_list[i].dest_ip);
      return ip_sticky_list[i].dest_ip;
    }
  }
  return 0;
}

void clear_rule_fld_values()
{
 int i = 0;
 for(i=0; i < num_rule_fields;i++)
   rule_fields[i].value[0] = '\0';

 return;
}

__u32 route_to(__u32 server_addr, int sticky_flag, __u32 saddr)
{
  __u32 dest_addr = 0;
 
      if(sticky_flag==STICKY_ON_IP) {
	/*printk("###### SERVER_ROUTE_TO=  STICKY  \n");*/
	ip_sticky_list[sticky_num].dest_ip=server_addr;
        ip_sticky_list[sticky_num].source_ip = saddr;
	sticky_num++;
	/*printk("##### in route_to sticky_num=%d \n",sticky_num);*/
      }
      dest_addr=server_addr;
      return dest_addr;
  return dest_addr;
}

u_long ruleMatching(char *url, char *method, u_long clientaddr, u_short clientport, char *content_type, char *body, int content_length)
{
  u_long ret_addr = 0;
  int ret=0,i=0;
  
  if((ret_addr = check_sticky(clientaddr)) > 0)
      goto return_directly;

  if(strncmp(method, "POST", 4) == 0)
  {
//    if(strstr(rule->content_type,"x-www-form-urlencoded") != NULL)
        printk(" encoded body when decoded =>%d\n",(ret=xml_tag_parser(strDecode(body), content_length)));

  }

  if(ret >= 0)
  {
     ret_addr = (u_long)  rule_configure(url,clientaddr, (u_short) clientport);
     clear_rule_fld_values();
  }

return_directly:
  return ret_addr;
}

u_long rule_configure(char *url, u_long saddr, u_short sport)
{
	int i;

    for(i=0;i < num_rule_fields;i++)
      printk("rule %s value %d\n ", rule_fields[i].name, atoi(rule_fields[i].value));

if (strstr(url,"gif") != NULL) {
	printk("match found with rule 1\n");
  return route_to(0x80c63cbc,NON_STICKY,saddr);
}

if (strstr(url,"cs522") != NULL) {
	printk("match found with rule 2\n");
  return route_to(0x80c63cb7,NON_STICKY,saddr);
}

if (strstr(url,"cs301") != NULL) {
	printk("match found with rule 3\n");
  return route_to(0x80c63cbc,NON_STICKY,saddr);
}

      printk("rule value %d\n ", ((atoi(rule_fields[1].value) > 0) && (atoi(rule_fields[1].value) <50000)));
      
if ((atoi(rule_fields[1].value) > 0) && (atoi(rule_fields[1].value) <50000)){
	printk("match found with rule 4\n");
  return route_to(0x80c63cb7,NON_STICKY,saddr);
}

if (atoi(rule_fields[1].value) > 50000) {
	printk("match found with rule 5\n");
  return route_to(0x80c63cbc,NON_STICKY,saddr);
}

	printk("match found with default rule \n");
  return route_to(0x80c63cbc,NON_STICKY,saddr);

}


char *mstrstr(char *source,  char *substr, size_t n)
{
  char *p = NULL;
  int  subStrLen = 0;

  p = strstr(source, substr);

  if(p == NULL)
  {
    subStrLen =  strlen(substr);
    if(!memcmp(source + n - subStrLen, substr, subStrLen))
      p = source + n - subStrLen;
  }
  return p;
}

int xml_tag_parser(char *buf, int  buf_len)
{
  int   ret = 0;
  int   field_count = 0;
  short done_yet = 0;
  int   skip_len = 0;
  int   pos = 0;
  char  tag[64];
  char  pos_str[10];
  char  *p = NULL;
  char  *p1 = NULL;
  char  *fld_spec_beg = NULL;
  char  *fld_spec_end = NULL;
  char  *data_beg = NULL;
  char  *data_end = NULL;
  char  *tag_beg = NULL;
  char  *tag_end = NULL;

  if (buf == NULL)
	  return -1;

  for(field_count = 0; field_count < num_rule_fields; field_count++)
  { 
    fld_spec_beg = &(rule_fields[field_count].name[0]);
    fld_spec_end = &(rule_fields[field_count].name[strlen(rule_fields[field_count].name)]);
    data_beg = buf;
    data_end =(char *) (buf+buf_len);

    while(fld_spec_beg < fld_spec_end)
    {
      p = (char *)memchr(fld_spec_beg, ':', fld_spec_end - fld_spec_beg);

      if(!p)
      {
        printk("Invalid rule field specification '%s'\n", rule_fields[field_count].name);
        return -1;
      }

      memcpy(tag, fld_spec_beg, p - fld_spec_beg);
      tag[p - fld_spec_beg] = '\0';

      p += 1;
      p1 = (char *)memchr(p, '.', fld_spec_end - p);


      if(!p1)
      {
        printk("Invalid rule field specification '%s'\n", rule_fields[field_count].name);
        return -1;
      }

      memcpy(pos_str, p, p1 - p);

      pos_str[p1 - p] = '\0';
      fld_spec_beg = ++p1;

      pos = atoi(pos_str); 
    
      get_tag_range(tag, pos, data_beg, data_end, &tag_beg, &tag_end);
      if((!tag_beg) || (!tag_end))
      {
              printk("get_tag_range() returned NULL either for tag_beg or tag_end, tag_beg=%u, tag_end=%u, tag=%s, pos=%d ",
                ((tag_beg) ? (unsigned int) tag_beg : 0),
                ((tag_end) ? (unsigned int) tag_end : 0),
                tag,
                pos);

        return -1;
      }

      if(fld_spec_beg == fld_spec_end)
      {
        /* We have parsed all the sub fields of the field specified in the rule.
        ** Now parse data between tag_beg and tag_end and store in the field value
        ** after skipping white spaces, carriage returns and line feeds.
        */
        done_yet = 0;
        while((tag_beg < tag_end) && !done_yet)
        {
          done_yet = 1;
          skip_len = len_sp;
          if(!memcmp(tag_beg, c_sp, skip_len))
          {
            tag_beg += skip_len;
            done_yet = 0;
          }

          skip_len = len_cr;

          if(!memcmp(tag_beg, c_cr, skip_len))
          {
            tag_beg += skip_len;
            done_yet = 0;
          }

          skip_len = len_lf;
          if(!memcmp(tag_beg, c_lf, skip_len))
          {
            tag_beg += skip_len;
            done_yet = 0;
          }
        }

        done_yet = 0;

        while((tag_beg < tag_end) && !done_yet)
        {
          skip_len = len_sp;
          if(memcmp(tag_end - skip_len, c_sp, skip_len))
            skip_len = 0;

          if(!skip_len)
          {
            skip_len = len_cr;
            if(memcmp(tag_end - skip_len, c_cr, skip_len))
              skip_len = 0;
          }
         
          if(!skip_len)
          {
            skip_len = len_lf;
            if(memcmp(tag_end - skip_len, c_lf, skip_len))
              skip_len = 0;
          }
         
          if(skip_len)
            tag_end -= skip_len;
          else
            done_yet = 1;
        }

        if(tag_beg != tag_end)
        {
          if((tag_end - tag_beg) < sizeof(rule_fields[field_count].value))
          {
            memcpy(rule_fields[field_count].value, tag_beg, tag_end - tag_beg);
            rule_fields[field_count].value[tag_end - tag_beg] = '\0';
          }
          else
            printk("tag value size=%d is larger than its container size=%d for rule tag '%s' of field spec '%s'\n", (tag_end - tag_beg), sizeof(rule_fields[field_count].value), tag, rule_fields[field_count].name);
        }
        else
          printk("No value found for rule field spec '%s'\n", rule_fields[field_count].name);
      }
      else
      {
        data_beg = tag_beg;
        data_end = tag_end;
      }
    }
  } 
  return ret;
}

int get_tag_range(char *tag, int pos, char *hay_begin, char *hay_end, char **needle_begin, char **needle_end)
{
  char  temp_buf[67];
  int   curr_pos = 0;
  int   skip_len = 0;
  int   hay_len = 0;
  int   buf_len = 0;
  char  *p = NULL;
  char  *p_tmp = NULL;
  char  *p1 = NULL;
  char  *p_end = NULL;
  char  *p_beg = NULL;
  char  *tag_left = NULL;
  char  *tag_right = NULL;
  
  hay_len = hay_end - hay_begin;
  buf_len = hay_len;
  p = hay_begin;

  while(curr_pos < pos)
  {
    /*
    ** First Remove any leading comments
    */
    sprintf(temp_buf, "%s%s--", c_la, c_ex);


    if((p_tmp = mstrstr(p, temp_buf, hay_len)) != NULL)
    {
      /*
      ** Found the beginning of comment marker "<!--"
      ** Now find the end of comment marker "-->"
      */
       
	sprintf(temp_buf, "--%s", c_ra);


       if((p1 = mstrstr(p_tmp, temp_buf, buf_len - (p_tmp - hay_begin))) != NULL)
       {
         p = p1 + strlen(temp_buf);
         hay_len = buf_len - (p - hay_begin);
         continue;
       }
       else
       {
         printk("Bad xml data, no ending comment tag\n");
         break;
       }

    }

    sprintf(temp_buf, "%s%s", c_fs, tag);
    /*
    ** First look for end position "/tag" i.e., the vicinity of </tag> 
    */

    if((p_end = mstrstr(p, temp_buf, hay_len)) != NULL)
    {
      curr_pos++;
      if(curr_pos == pos)
      {
        /*
        ** Now look for begin position of the "tag" i.e., the vicinity of <tag>
        */

        if((p_beg = mstrstr(p, tag, hay_len)) != NULL)
        {
          /*
          ** First verify that it is a real tag 
          */

           tag_left = p_beg;
           tag_right = p_beg + strlen(tag);

          /*
          ** i.e, Below is a check for situations like <sometag> AtagB </sometag>
          */

           if(( (memcmp(tag_left - len_sp, c_sp, len_sp)) &&
                (memcmp(tag_left - len_ra, c_la, len_la)) &&
                (memcmp(tag_left - len_cr, c_cr, len_cr)) &&
                (memcmp(tag_left - len_lf, c_lf, len_lf)) ) || 
              ( (memcmp(tag_right, c_sp, len_sp)) &&
                (memcmp(tag_right, c_ra, len_ra)) &&
                (memcmp(tag_right, c_cr, len_cr)) &&
                (memcmp(tag_right, c_lf, len_lf)) ) )
           {
             p = tag_right;
             hay_len = buf_len - (p - hay_begin);
             --curr_pos;
             continue;
           }

           p_beg = tag_right;

           /*
           ** Find the position of '>' for p_beg
           */
           p = mstrstr(p_beg, c_ra, p_end - p_beg);

           if(p == NULL)
           {
             p_beg = NULL;
             p_end = NULL;
             break;
           }

           /*
           ** Find the position of '<' for beginning from p_beg
           */
           p_tmp = mstrstr(p_beg, c_la, p_end - p_beg);

           /*
           ** Below is a check for <sometag> tag </sometag>
           */

           if(p_tmp && (p_tmp < p))
           {
             hay_len = buf_len - (p - hay_begin);
             --curr_pos;
             continue;
           }

          /*
          ** Now eliminate all unwanted characters between p_beg and p_end
          ** such as white space, line feed, carriage return until until
          ** right and left anchors are seen respectively
          */

           if(p != NULL)
             p_beg = p + len_ra;
           else
           {
             /*
             ** We didn't find the  '>' for p_beg
             */
             p_beg = NULL;
             p_end = NULL;
             break;
           }

           /*
           ** Find the position of left anchor for p_end
           */

           while((p_end - max_char_len) > p_beg)
           {
             skip_len = len_la;
             if(!memcmp(p_end - skip_len, c_la, skip_len))
             {
               p_end -= skip_len;
               break;
             }

             skip_len = len_sp;
             if(!memcmp(p_end - skip_len, c_sp, skip_len))
               p_end -= skip_len;


             skip_len = len_cr;
             if(!memcmp(p_end - skip_len, c_cr, skip_len))
               p_end -= skip_len;

             skip_len = len_lf;
             if(!memcmp(p_end - skip_len, c_lf, skip_len))
               p_end -= skip_len;
           }
        }
      }
      else
      {
        /*
        ** Advance p for the next <tag>...data...</tag> block
        */
        p = p_end + strlen(temp_buf);
        hay_len = buf_len - (p - hay_begin);
      }
    }
    else
    /*
    ** No specified needle found in the hay
    */
      break;

  }

  *needle_begin = p_beg;
  *needle_end = p_end;
  
 return 0;

}

/*
** borrwed from thttpd
*/

static int hexDigit(char c)
{
    if (strchr("0123456789abcdefABCDEF",c) != (char *) NULL)
        return (1);

    return (0);
}

/*
** from thttpd source code
*/
static int hexIt(char c)
{
    if ((c >= '0') && (c <= '9'))
        return (c-'0');

    if ((c >= 'a') && (c <= 'f'))
        return (c-'a'+10);

    if ((c >= 'A') && (c <= 'F'))
        return (c-'A'+10);

    return (0);           /* shouldn't happen, but for saftey  */

}


/*
** from thttpd source code
*/
char *strDecode(char *from)
{
char *to;
char *start=NULL;

    if (from  == (char *) NULL)
        return ((char *) NULL);


    to = (char *) kmalloc ((int) strlen(from) + 1, GFP_KERNEL);
    
    if (to == (char *) NULL)
    {
        printk("memory allocation problem strDecode()");
        return NULL;
    }

    start = to;

    for (; *from != '\0'; ++to,++from)
    {
        if (*from == '%' && hexDigit(from[1]) &&
            hexDigit(from[2]))
        {
            *to=hexIt(from[1])*16+hexIt(from[2]);
            from += 2;
        }
        else
            *to = (*from=='+') ?' ': *from;      /* inorder to replace + in post request with a space */
    }
    *to='\0';

return start;
}



/***************************END RULE MODULE**************************/

static int tcp_vs_lcsphttp_init_svc(struct tcp_vs_service *svc)
{
	        return 0;
}


static int tcp_vs_lcsphttp_done_svc(struct tcp_vs_service *svc)
{
	        return 0;
}


static int tcp_vs_lcsphttp_update_svc(struct tcp_vs_service *svc)
{
	        return 0;
}


/****************************************************************************
*	This doesn't accept 0x if the radix is 16. The overflow code assumes
*	a 2's complement architecture
*/
#ifndef strtol
static long strtol( char *string, char **endptr, int radix)
{
	char *s;
	long value;
	long new_value;
	int sign;
	int increment;

	value = 0;
	sign = 1;
	s = string;

	if ((radix == 1) || (radix > 36) || (radix < 0)) {
		goto done;
	}

	/* skip whitespace */
	while ((*s == ' ') || (*s == '\t') || (*s == '\n') || (*s == '\r')) {
		s++;
	}

	if (*s == '-') {
		sign = -1;
		s++;
	}
	else if (*s == '+') {
		s++;
	}

	if (radix == 0) {
		if (*s == '0') {
			s++;
			if ((*s == 'x') || (*s == 'X')) {
				s++;
				radix = 16;
			}
			else
				radix = 8;
		}
		else
			radix = 10;
	}

	/* read number */
	while (1) {
		if ((*s >= '0') && (*s <= '9'))
			increment = *s - '0';
		else if ((*s >= 'a') && (*s <= 'z'))
			increment = *s - 'a' + 10;
		else if ((*s >= 'A') && (*s <= 'Z'))
			increment = *s - 'A' + 10;
		else
			break;

		if (increment >= radix)
			break;

		new_value = value * radix + increment;
		/* detect overflow */
		if ((new_value - increment)/radix != value) {
			s = string;
			value = -1 >> 1;
			if (sign < 0)
				value += 1;

			goto done;
		}

		value = new_value;
		s++;
	}

  done:
	if (endptr)
		*endptr = s;

	return value*sign;
}
#endif


/****************************************************************************
*	relay data between source socket and destination socket
*/
static int get_http_body(char *body,	/* packet body */
	http_read_ctl_block_t *ctl_blk,	/* read control block with source socket */
	int len				/* relay data length */
	)
{
	int nbytes, reads, w = 0;
	int ret = -1;

	DECLARE_WAIT_QUEUE_HEAD(WQ);

	EnterFunction(5);

	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset))
	assert(len > 0)

	/* if there is enough data in read buffer */
	nbytes = len - ctl_blk->remaining;
	if (nbytes <= 0) 
	{
		strncpy(body, ctl_blk->buffer + ctl_blk->offset, len);
	
		ctl_blk->offset		+= len;
		ctl_blk->remaining	-= len;
		goto done;
	}

	/* copy the remaining bytes */
	if (ctl_blk->remaining > 0) {
		strncat(body, ctl_blk->buffer + ctl_blk->offset, ctl_blk->remaining);
	}

	do {
		reads = tcp_vs_recvbuffer (ctl_blk->sock, ctl_blk->buffer, ctl_blk->len, ctl_blk->flag);
		if (reads == 0) {
			interruptible_sleep_on_timeout(&WQ, HZ);
			TCP_VS_DBG (5, "Reads 0 bytes while relay\n");
			continue;
		}

		if (reads < 0) {
			TCP_VS_ERR("Error in reading while relaying\n");
			goto exit;
		}

		w = MIN(nbytes, reads);
		strncat(body, ctl_blk->buffer, w);

		nbytes -= w;
	}while (nbytes > 0);

	ctl_blk->offset		= w;
	ctl_blk->remaining	= reads - w;

	assert(ctl_blk->remaining >= 0)
	assert(ctl_blk->offset  < ctl_blk->len)
done:
	ret = 0;
exit:
	LeaveFunction(5);
	return ret;
}


/****************************************************************************
*	skip whitespace
*/
static inline char * skip_lws (const char* buffer)
{
	char* s = (char*)buffer;
	while ((*s == ' ') || (*s == '\t') || (*s == '\n') || (*s == '\r')) {
		s++;
	}
	return s;
}


/****************************************************************************
*   Parse http request line. (request line is terminated by CRLF)
*
*   RFC 2616, 19.3
*   Clients SHOULD be tolerant in parsing the Status-Line and servers
*   tolerant when parsing the Request-Line. In particular, they SHOULD
*   accept any amount of SP or HT characters between fields, even though
*   only a single SP is required.
*
*/
static int parse_http_request_line (char *buffer, size_t len, http_request_t *req)
{
	char	*pos, c;
	int	ret = PARSE_ERROR;
	int	i;

	EnterFunction(5);

	/* terminate string */
	c = buffer[len];
	buffer[len] = 0;

	TCP_VS_DBG(5, "parsing request:\n");
	TCP_VS_DBG(5, "--------------------\n");
	TCP_VS_DBG(5, "%s\n", buffer);
	TCP_VS_DBG(5, "--------------------\n");

	req->message		= buffer;
	req->message_len	= len;

	/*
	 * RFC 2616, 5.1:
	 *	Request-Line = Method SP Request-URI SP HTTP-Version CRLF
	 */

	/* try to get method */
	pos = skip_lws (buffer);
	req->method = HTTP_M_UNKNOWN;
	for (i = 1; i < HTTP_M_MAX; i++) {
		if (strnicmp (pos, http_methods[i].name, http_methods[i].len) == 0) {
			req->method = i;
			break;
		}
	}
	if (req->method == HTTP_M_UNKNOWN) {
		goto exit;
	}
	TCP_VS_DBG(6, "HTTP METHOD: %s\n", http_methods[i].name);

	pos += http_methods[i].len;

	/* get URI string */
	req->uri_str = skip_lws (pos + 1);
	TCP_VS_DBG(6, "URI: %s\n", req->uri_str);

	if ((pos = strchr ((char*)req->uri_str, SP)) == NULL) {
		goto exit;
	}

	req->uri_len = pos - req->uri_str;

	/* get http version */
	req->version_str	= skip_lws (pos + 1);
	req->version	= HTTP_V_UNKNOWN;
	for (i = 0; http_protocol_strings[i] != NULL; i++) {
		if (strnicmp (req->version_str, http_protocol_strings[i],
			      strlen (http_protocol_strings[i])) == 0) {
			req->version = i + HTTP_V_0_9;
			break;
		}
	}

	if (req->version == HTTP_V_UNKNOWN) {
		goto exit;
	}

	TCP_VS_DBG(6, "HTTP VERSION: %s\n", http_protocol_strings[i]);
	ret = PARSE_OK;
  exit:
	buffer[len] = c;	/* restore string */
	LeaveFunction(5);
	return ret;
}

/****************************************************************************
*	Is there any data in socket?
*
* return:
*	-1,	Socket error
*	 0,	No data can read from socket
*	 1,	Data available
*/
static inline int data_available (http_read_ctl_block_t *ctl_blk)
{
	int ret;

	EnterFunction(12);

	if (ctl_blk->remaining == 0) {
		/* check if the connection is closed */
		if (ctl_blk->sock->sk->state != TCP_ESTABLISHED
		    && ctl_blk->sock->sk->state != TCP_CLOSE_WAIT) {
			ret = -1;
			goto out;
		}

		/* Do we have data ? */
		if (skb_queue_empty(&(ctl_blk->sock->sk->receive_queue)))
			ret = 0;
		else
			ret = 1;
		goto out;
	} else
		ret = 1;

  out:
	LeaveFunction(12);
	return ret;
}

/****************************************************************************
*
* http_read_line - read a line from socket.
*
* Try to get the line from the remaining bytes. then read max_line_size bytes
* from socket and get a line. (the line delimeter is "CRLF")
* Return the len of the line (not including CRLF), or -1 if failed.
*
*   Note: 1, http_read_line does not terminate the line with '\0', it still end
*   with CRLF.
*	2, the acutul search length may bigger than max_line_size.
*/
static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
{
	char* buf;
	int nbytes, i, offset, reads;
	int len = -1;

	DECLARE_WAIT_QUEUE_HEAD(WQ);

	EnterFunction(5);

	assert(max_line_size < ctl_blk->len);
	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));

	ctl_blk->info = NULL;
	if (ctl_blk->remaining == 0) {
		ctl_blk->offset = 0;
	}

	offset = ctl_blk->offset;
	buf = ctl_blk->buffer + offset;

	/* try to get a line from the remaining bytes */
	for (i = 0; i < ctl_blk->remaining -1; i++) {
		if ((buf[i] == CR) && (buf[i + 1] == LF)) {
			len = i;
			goto done;
		}
	}
	/* overflow? */
	if ((offset + max_line_size) > ctl_blk->len) {
		memmove (ctl_blk->buffer, buf, ctl_blk->remaining);
		ctl_blk->offset = 0;
		buf = ctl_blk->buffer;
	}

	nbytes = max_line_size - ctl_blk->remaining;

	/* try to read a line from the socket */
	while ((nbytes > 0) && (len < 0)) {
		/* go out if the connection is closed */
		if (ctl_blk->sock->sk->state != TCP_ESTABLISHED
		    && ctl_blk->sock->sk->state != TCP_CLOSE_WAIT) {
			if (len > 0)
				break;
			else
				goto exit;
		}

		assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));
		reads = tcp_vs_recvbuffer(ctl_blk->sock,
					  ctl_blk->buffer + ctl_blk->offset + ctl_blk->remaining,
					  ctl_blk->len - ctl_blk->offset - ctl_blk->remaining,
					  ctl_blk->flag);

		if (reads == 0) {
			interruptible_sleep_on_timeout(&WQ, HZ);
			TCP_VS_DBG(5, "Read 0 bytes while reading a line\n");
			continue;
		}

		if (reads < 0) {
			TCP_VS_ERR("Error in reading a line\n");
			goto exit;
		}

		ctl_blk->remaining += reads;

		/* try to get a line from the remaing bytes */
		for (; i < ctl_blk->remaining -1; i++) {
			if ((buf[i] == CR) && (buf[i + 1] == LF)) {
				len = i;
				goto done;
			}
		}
		nbytes -= reads;
	}
  done:
	ctl_blk->info		= buf;
	ctl_blk->offset		+= len + 2;
	ctl_blk->remaining	-= len + 2;

	assert(ctl_blk->remaining >= 0);
	assert(ctl_blk->offset  < ctl_blk->len);

	printk("length == %d\n", len+2);
  exit:
	LeaveFunction(5);
	return len;
}


/****************************************************************************
* http_line_unescape - convert escaped characters in buffer to ASCII
*
* This routine can be used to convert an "escaped" form of a URL or
* parameter (appended to a URL) to standard ASCII format.
* The escaping is done by the browser on the client side, for
* transferring characters not allowed by the HTTP protocol.
* For example, a whitespace character is not allowed in an URL.
* It must be substituted by an escape sequence to be transferred.
*
* ESCAPING
* When you want to include any character, not part of the standard
* set allowed in URLs, you can do this by specifying its hex value
* in the format %xx, where xx is the hex representation.
* In addition, every '+' character will be substituted by a space.
*
*/
static void http_line_unescape
	(
	char	*string,	/* escaped string to unescape */
	int	len	/* length of the string */
	)
{
	int	i = 0;
	char	buffer[3];
	char	c;

	EnterFunction(5);

	assert(string != NULL)

	while (i < len) {
		if (string[i] == '+') {
			string[i] = ' ';	/* replace '+' by spaces */
		}
		if ((string[i] == '%') && (i < len-2)) {
			if (isxdigit (string[i+1]) && isxdigit (string[i+2])) {
				strncpy (buffer, &(string[i+1]), 2);
				buffer[2] = 0;
				c = (char) strtol (buffer, NULL, 16);
				if (c != 0) {
					memmove (&(string [i]),	/* move string 2 chars */
						&(string[i+2]), 2);
					string[i] = c;	/* replace % by new char */
					len -= 2;
				}
			}
		}
		i++;
	}

	LeaveFunction(5);
	return;
}


/******************************************************************************
* http_mime_parse - parse MIME line in a buffer
*
* This routine parses the MIME line in a buffer.
*
* NOTE: Some MIME headers (host, Referer) need be considered again, tbd.
*
*/
static int http_mime_parse
	(
	char			*buffer,
	int			len,
	http_mime_header_t	*mime,
	int			partial
	)
{
	char	*pos, *buf, c;
	int	l, ret = PARSE_ERROR;

	EnterFunction(5);

	assert(buffer != NULL)

		/* terminate string */
		c = buffer[len];
	buffer[len] = 0;

	buffer = skip_lws (buffer);
	if ((pos = strchr (buffer, ':')) == NULL) {
		goto exit;
	}

	l = pos - buffer;

	TCP_VS_DBG(5, "MIME Header: %s\n", buffer);

	/*
	 * Messages MUST NOT include both a Content-Length header field and
	 * a non-identity transfer-coding. If the message does include a non-
	 * identity transfer-coding, the Content-Length MUST be ignored.
	 */
	if (strnicmp (buffer, "Transfer-Encoding", l) == 0) { /* maybe strlen is better than 'l' */
		pos = skip_lws (pos + 1);
		if (strnicmp (pos, "identity", strlen ("identity")) != 0) {
			mime->transfer_encoding = 1;
			TCP_VS_DBG(5, "Transfer-Encoding: chunked\n");
		}
	}
	else if (strnicmp (buffer, "Content-Length", l) == 0) {
		mime->content_length = strtol (pos + 1, NULL, 10);
		TCP_VS_DBG(5, "Content-Length: %d\n", mime->content_length);
	} else if (strnicmp (buffer, "Connection", l) == 0) {
		pos = skip_lws (pos + 1);
		if (strnicmp (pos, "close", strlen ("close")) == 0) {
			mime->connection_close = 1;
			TCP_VS_DBG(5, "Connection: close\n");
		}
	}
	else if (partial && (strnicmp (buffer, "Content-type", l) == 0)){
		l = strlen ("multipart/byteranges");
		pos = skip_lws (pos + 1);
		if (strnicmp (pos, "multipart/byteranges", l)== 0){
			TCP_VS_DBG(5, "multipart/byteranges\n");
			pos += l + 1; /* skip ';' */
			pos = skip_lws (pos + 1);
			l = strlen ("boundary=");
			if (strnicmp (pos, "boundary=", l) != 0) {
				goto exit;
			}
			buf = pos + l;

			/* the rest of this line is THIS_STRING_SEPARATES */
			l = buffer + len - buf;
			if ((mime->sep = kmalloc (l + 1, GFP_KERNEL)) == NULL) {
				goto exit;
			}

			/* RFC 2046 [40] permits the boundary string to be quoted */
			if (buf[0] == '"' || buf[0] == '\'') {
				buf++;
				l--;
			}
			strncpy (mime->sep, buf, l);
			mime->sep[l] = 0;
			TCP_VS_DBG(5, "THIS_STRING_SEPARATES : %s\n", mime->sep);
		}else {
			TCP_VS_DBG(5, "Content-type : %s\n", pos);
		}
			
			
		
	}

	ret = PARSE_OK;
  exit:
	buffer[len] = c;	/* restore the string */
	LeaveFunction(5);
	return ret;
}


static inline struct tcp_vs_dest *
__tcp_vs_phttp_wlc_schedule(struct list_head *destinations)
{
	register struct list_head *e;
	struct tcp_vs_dest *dest, *least;

	list_for_each (e, destinations) {
		least = list_entry(e, struct tcp_vs_dest, r_list);
		if (least->weight > 0) {
			goto nextstage;
		}
	}
	return NULL;

	/*
	 *	Find the destination with the least load.
	 */
  nextstage:
	for (e=e->next; e!=destinations; e=e->next) {
		dest = list_entry(e, struct tcp_vs_dest, r_list);
		if (atomic_read(&least->conns) * dest->weight >
		    atomic_read(&dest->conns) * least->weight) {
			least = dest;
		}
	}

	return least;
}


static struct tcp_vs_dest * tcp_vs_phttp_matchrule(struct tcp_vs_service *svc, http_request_t *req, struct tcp_vs_conn *conn, char *content_type, char *body)
{

	register struct list_head *l, *e;
	struct tcp_vs_dest *least = NULL;
	u_long dest;

	 char *uri;

         if (!(uri = kmalloc(req->uri_len+1, GFP_KERNEL))) {
                 TCP_VS_ERR("No memory!\n");
                 return NULL;
         }

	 memcpy(uri, req->uri_str, req->uri_len);
	 uri[req->uri_len] = '\0';
         TCP_VS_DBG(5, "matching request URI: %s method %s\n", uri, http_methods[req->method].name);

     dest = ruleMatching( uri, http_methods[req->method].name, conn->addr,conn->port, content_type, body,req->mime.content_length);
		
	read_lock(&svc->lock);

	l = &svc->destinations;
	for (e=l->next; e!=l; e=e->next) {
	  least = list_entry(e, struct tcp_vs_dest, n_list);
	       if (ntohl(least->addr) == dest)
		     break;
	}
	
      printk("address yahoo least %x dest %x\n", ntohl(least->addr), dest);
	
      read_unlock(&svc->lock);
	
	return least;
}


#if 0
u_long route_to(char *server_name, int sticky_flag, u_long saddr)
{
  int i=0;
  u_long dest_addr = 0;

  for(i=0; i<num_servers; i++)
     {
      if(strcmp(server_name, server_list[i].name) == 0)
        {
	    if(sticky_flag == STICKY_ON_IP)
	      {
		 addSticky(saddr,server_list[i].addr);
              }
	    dest_addr = server_list[i].addr;
	    break;
	}
     }
   return dest_addr;
}

#endif







/****************************************************************************
*  Parse a chunk extension, detect overflow.
*  There are two error cases:
*  1) If the conversion would require too many bits, a -1 is returned.
*  2) If the conversion used the correct number of bits, but an overflow
*     caused only the sign bit to flip, then that negative number is
*     returned.
*  In general, any negative number can be considered an overflow error.
*/
static long get_chunk_size(char *b)
{
	long chunksize = 0;
	size_t chunkbits = sizeof(long) * 8;

	/* skip whitespace */
	while ((*b == ' ') || (*b == '\t') || (*b == '\n') || (*b == '\r')) {
		b++;
	}

	/* Skip leading zeros */
	while (*b == '0') {
		++b;
	}

	while (isxdigit(*b) && (chunkbits > 0)) {
		int xvalue = 0;

		if (*b >= '0' && *b <= '9') {
			xvalue = *b - '0';
		}
		else if (*b >= 'A' && *b <= 'F') {
			xvalue = *b - 'A' + 0xa;
		}
		else if (*b >= 'a' && *b <= 'f') {
			xvalue = *b - 'a' + 0xa;
		}

		chunksize = (chunksize << 4) | xvalue;
		chunkbits -= 4;
		++b;
	}
	if (isxdigit(*b) && (chunkbits <= 0)) {
		/* overflow */
		return -1;
	}

	return chunksize;
}


#if 0
int forward_http_message(struct sock *sock, char *header, int header_len, char *body, http_mime_header_t *mime)
{
	int ret = 1;
	if (tcp_vs_xmit (sock, header, header_len , 0) < 0) {
		TCP_VS_ERR("Error in xmitting trailer\n");
		ret = 0;
		goto exit;
	}

	if (tcp_vs_xmit (sock, body, mime->content_length,0) < 0) {
		TCP_VS_ERR("Error in xmitting trailer\n");
		ret = 0;
		goto exit;
	}
	
	return ret;
	
exit:
	return ret;
}
#endif


/****************************************************************************
* parse_http_status_line - parse the http status line.
*
*   RFC 2616, 19.3
*   Clients SHOULD be tolerant in parsing the Status-Line and servers
*   tolerant when parsing the Request-Line. In particular, they SHOULD
*   accept any amount of SP or HT characters between fields, even though
*   only a single SP is required.
*
*/
static int parse_http_status_line (char *buffer, size_t len, http_response_t *resp)
{
	char *pos, c;
	int i, ret = PARSE_ERROR;

	EnterFunction(5);

	assert(buffer != NULL)

	/* terminate string */
	c = buffer[len];
	buffer[len] = '\0';

	TCP_VS_DBG(5, "parsing response:\n");
	TCP_VS_DBG(5, "--------------------\n");
	TCP_VS_DBG(5, "%s\n", buffer);
	TCP_VS_DBG(5, "--------------------\n");

	/*
	 * RFC 2616, 6.1:
	 *	Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF
	 */

	pos = skip_lws (buffer);

	resp->version = HTTP_V_UNKNOWN;
	for (i = 0; http_protocol_strings[i] != NULL; i++) {
		if (strnicmp (pos, http_protocol_strings[i],
			strlen (http_protocol_strings[i])) == 0) {
			resp->version = i + HTTP_V_0_9;
			break;
		}
	}

	if (resp->version != HTTP_V_UNKNOWN) {
		/* get the status code */
		pos += strlen (http_protocol_strings[i]) + 1;
		resp->status_code = strtol (pos, NULL, 10);
		assert(resp->status_code >= 100)
		TCP_VS_DBG(6, "Status Code: %d\n", resp->status_code);
		ret = PARSE_OK;
	}

	buffer[len] = c;	/* restore string */
	LeaveFunction(5);
	return ret;
}


/****************************************************************************
*	relay data between source socket and destination socket
*/
static int relay_http_data
	(
	struct socket* dsock,		/* destination socket */
	http_read_ctl_block_t *ctl_blk,	/* read control block with source socket */
	int len				/* relay data length */
	)
{
	int nbytes, reads, w = 0;
	int ret = -1;

	DECLARE_WAIT_QUEUE_HEAD(WQ);

	EnterFunction(5);

	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset))
	assert(len > 0)

	/* if there is enough data in read buffer */
	nbytes = len - ctl_blk->remaining;
	if (nbytes <= 0) {
		if (tcp_vs_xmit (dsock, ctl_blk->buffer + ctl_blk->offset,
			len, MSG_MORE) < 0) {
			TCP_VS_ERR("Error in xmitting message body\n");
			goto exit;
		}
		ctl_blk->offset		+= len;
		ctl_blk->remaining	-= len;
		goto done;
	}

	/* xmit the remaining bytes */
	if (ctl_blk->remaining > 0) {
		if (tcp_vs_xmit (dsock, ctl_blk->buffer + ctl_blk->offset,
			ctl_blk->remaining, MSG_MORE) < 0) {
				TCP_VS_ERR("Error in xmitting remaining bytes\n");
				goto exit;
		}
	}

	do {
		reads = tcp_vs_recvbuffer (ctl_blk->sock, ctl_blk->buffer, ctl_blk->len, ctl_blk->flag);
		if (reads == 0) {
			interruptible_sleep_on_timeout(&WQ, HZ);
			TCP_VS_DBG (5, "Reads 0 bytes while relay\n");
			continue;
		}

		if (reads < 0) {
			TCP_VS_ERR("Error in reading while relaying\n");
			goto exit;
		}

		w = MIN(nbytes, reads);

		if (tcp_vs_xmit (dsock, ctl_blk->buffer, w, MSG_MORE) < 0) {
			TCP_VS_ERR("Error in relaying bytes\n");
			goto exit;
		}

		nbytes -= w;
	}while (nbytes > 0);

	ctl_blk->offset		= w;
	ctl_blk->remaining	= reads - w;

	assert(ctl_blk->remaining >= 0)
	assert(ctl_blk->offset  < ctl_blk->len)
done:
	ret = 0;
exit:
	LeaveFunction(5);
	return ret;
}

/****************************************************************************
*	search the seperator in a string
*/
static char* search_sep (const char * s, int len, const char * sep)
{
	int l, ll;

	l = strlen(sep);
	if (!l)
		return (char *)s;

	ll = len;
	while (ll >= l) {
		ll--;
		if (!memcmp(s, sep, l))
			return (char *) s;
		s++;
	}
	return NULL;
}


/****************************************************************************
* relay_multiparts: relay multipart/byteranges body
*
*  relay all data until "CRLF--THIS_STRING_SEPARATES--CRLF" is found.
*
*  Note: there may be a endless loop if the separate string is not found, tbd.
*
*/
static int relay_multiparts
	(
	struct socket* dsock,
	http_read_ctl_block_t *ctl_blk,
	http_mime_header_t *mime
	)
{
	int len, sep_len, l, reads;
	int ret = -1;
	char *buf, *pos;
	char *sep = NULL;

	DECLARE_WAIT_QUEUE_HEAD(WQ);

	EnterFunction(5);

	sep_len = strlen (mime->sep) + 8;
	if ((sep = kmalloc (sep_len + 1, GFP_KERNEL)) == NULL) {
		goto exit;
	}

	snprintf (sep, sep_len + 1, "\r\n--%s--\r\n" , mime->sep);

	/* deal with the remaining bytes */
	buf = ctl_blk->buffer + ctl_blk->offset;
	len = ctl_blk->remaining;

	if ((len > 0) && (tcp_vs_xmit (dsock, buf, len, MSG_MORE) < 0)) {
		TCP_VS_ERR("Error in xmitting multiparts (remaining)\n");
		goto exit;
	}

	pos = search_sep (buf, len, sep);
	if (pos != NULL) {
		goto done;
	}

	l = MIN(len, sep_len);
	memmove (ctl_blk->buffer, buf + len - l, l);

	/* search for CRLF--THIS_STRING_SEPARATES--CRLF */
	while (1) {
		reads = tcp_vs_recvbuffer (ctl_blk->sock, ctl_blk->buffer + l, ctl_blk->len - l, 0);
		if (reads == 0) {
			interruptible_sleep_on_timeout(&WQ, HZ);
			TCP_VS_DBG (5, "Reads 0 bytes while relaying multiparts\n");
		}

		if (reads < 0) {
			TCP_VS_ERR("Error in receiving multiparts\n");
			goto exit;
		}

		if (tcp_vs_xmit (dsock, ctl_blk->buffer + l, reads, MSG_MORE) < 0) {
			TCP_VS_ERR("Error in xmitting multiparts\n");
			goto exit;
		}

		len = l + reads;
		pos = search_sep (ctl_blk->buffer, len, sep);
		if (pos != NULL) {
			goto done;
		}

		l = MIN(len, sep_len);
		memmove (ctl_blk->buffer, ctl_blk->buffer + len - l, l);
	}


done:
	ret = 0;
exit:
	if (sep) {
		kfree (sep);
	}

	LeaveFunction(5);
	return ret;
}

/****************************************************************************
* transfer http message body.
*
* When a message-body is included with a message, the transfer-length of that
* body is determined by one of the following (in order of precedence):
*  1. Any response message which "MUST NOT" include a message-body (such as
* the 1xx, 204, and 304 responses and any response to a HEAD request) is always
* terminated by the first empty line after the header fields, regardless of the
* entity-header fields present in the message.
*  2. If a Transfer-Encoding header field (section 14.41) is present and has
* any value other than "identity", then the transfer-length is defined by use of
* the "chunked" transfer-coding (section 3.6), unless the message is terminated
* by closing the connection.
*  3. If a Content-Length header field (section 14.13) is present, its decimal
* value in OCTETs represents both the entity-length and the transfer-length. The
* Content-Length header field MUST NOT be sent if these two lengths are different
* (i.e., if a Transfer-Encoding header field is present). If a message is received
* with both a Transfer-Encoding header field and a Content-Length header field,
* the latter MUST be ignored.
*  4. If the message uses the media type "multipart/byteranges", and the transfer-
* length is not otherwise specified, then this self-delimiting media type defines
* the transfer-length. This media type MUST NOT be used unless the sender knows
* that the recipient can arse it; the presence in a request of a Range header with
* multiple byte-range specifiers from a 1.1 client implies that the client can parse
* multipart/byteranges responses.
*      A range header might be forwarded by a 1.0 proxy that does not
	understand multipart/byteranges; in this case the server MUST
	delimit the message using methods defined in items 1,3 or 5 of this
	section.
*  5. By the server closing the connection. (Closing the connection cannot be
* used to indicate the end of a request body, since that would leave no possibility
* for the server to send back a response.)
*
*/
static int relay_http_message_body
	(
	struct socket* dsock,		/* destination socket */
	http_read_ctl_block_t *ctl_blk,	/* read control block */
	http_mime_header_t *mime
	)
{
	int ret = -1;

	EnterFunction(5);

	if (mime->transfer_encoding) {
		/*
		 * 19.4.6 Introduction of Transfer-Encoding
		 *           HTTP/1.1 introduces the Transfer-Encoding header field (section
		 *           14.41). Proxies/gateways MUST remove any transfer-coding prior to
		 *        forwarding a message via a MIME-compliant protocol.
		 *           A process for decoding the "chunked" transfer-coding (section 3.6) can be
		 *        represented in pseudo-code as:
		 *               length := 0
		 *               read chunk-size, chunk-extension (if any) and CRLF
		 *               while (chunk-size > 0) {
		 *                      read chunk-data and CRLF
		 *                      append chunk-data to entity-body
		 *                      length := length + chunk-size
		 *                      read chunk-size and CRLF
		 *               }
		 *               read entity-header
		 *               while (entity-header not empty) {
		 *                      append entity-header to existing header fields
		 *                      read entity-header
		 *               }
		 *               Content-Length := length
		 *               Remove "chunked" from Transfer-Encoding
		 */
		int len, chunk_size;
		do {
			len = http_read_line (ctl_blk, HTTP_MAX_MIMLINE_SIZE);
			if (len < 0) {
				TCP_VS_ERR("Error in reading chunk size from client\n");
				goto exit;
			}

			if (tcp_vs_xmit (dsock, ctl_blk->info, len + 2, MSG_MORE) < 0) {
				TCP_VS_ERR("Error in xmitting chunk size & extension\n");
				goto exit;
			}

			ctl_blk->info[len] = 0;
			chunk_size = get_chunk_size (ctl_blk->info);

			TCP_VS_DBG(5, "Chunked line: %s\n", ctl_blk->info);

			if (chunk_size > 0) {
				if (relay_http_data (dsock, ctl_blk, chunk_size + 2) < 0) {
					TCP_VS_ERR("Error in xmitting chunk data\n");
					goto exit;
				}
			}
		} while ( chunk_size > 0);

		/* relay the trailer */
		do {
			len = http_read_line (ctl_blk, HTTP_MAX_MIMLINE_SIZE);
			if (len < 0) {
				TCP_VS_ERR("Error in reading trailer.\n");
				goto exit;
			}
			if (tcp_vs_xmit (dsock, ctl_blk->info, len + 2, MSG_MORE) < 0) {
				TCP_VS_ERR("Error in xmitting trailer\n");
				goto exit;
			}
		} while (len != 0);
		ret = 0;
	} else if (mime->content_length) {
		ret = relay_http_data (dsock, ctl_blk, mime->content_length);
	} else if (mime->sep) {
		ret = relay_multiparts (dsock, ctl_blk, mime);
	} else {
		ret = 0;		/* ? */
	}

  exit:
	LeaveFunction(5);
	return ret;
}




/****************************************************************************
*	get response from the specified server
*/
static int
http_get_response (struct socket *csock, struct socket* dsock,
	http_request_t *req, char* buffer, int buflen, int *close)
{
	http_read_ctl_block_t	read_ctl_blk;
	http_response_t		resp;
	int			len, ret = -1;

	EnterFunction(5);

	memset (&read_ctl_blk, 0, sizeof (read_ctl_blk));
	read_ctl_blk.buffer	= buffer;
	read_ctl_blk.len	= buflen;
	read_ctl_blk.sock	= dsock;

	*close = 0;

	/* Do we have data ? */
	while (skb_queue_empty(&(dsock->sk->receive_queue))) {
		interruptible_sleep_on_timeout(&dsock->wait, HZ);
	}

	/* read status line from server */
	len = http_read_line (&read_ctl_blk, HTTP_MAX_STALINE_SIZE);
	if (len < 0) {
		TCP_VS_ERR("Error in reading status line from server\n");
		goto exit;
	}

	/* xmit status line to client (2 more bytes for CRLF) */
	if (tcp_vs_xmit (csock, read_ctl_blk.info, len + 2,
			 MSG_MORE) < 0) {
		TCP_VS_ERR("Error in sending status line\n");
		goto exit;
	}

	/* parse status line */
	memset (&resp, 0, sizeof(resp));
	if (parse_http_status_line (read_ctl_blk.info, len, &resp) == PARSE_ERROR) {
		goto exit;
	}

	/* parse MIME header */
	do {
		if ((len = http_read_line (&read_ctl_blk, HTTP_MAX_MIMLINE_SIZE)) < 0) {
			goto exit;
		}

		/* xmit MIME header (2 more bytes for CRLF) */
		if (tcp_vs_xmit (csock, read_ctl_blk.info, len + 2, MSG_MORE) < 0) {
			TCP_VS_ERR("Error in sending status line\n");
			goto exit;
		}

		http_line_unescape (read_ctl_blk.info, len);
		http_mime_parse (read_ctl_blk.info, len, &resp.mime, resp.status_code == 206);
	} while (len != 0); /* http header end with CRLF,CRLF */

	*close = resp.mime.connection_close;

	/*
	 * Any response message which "MUST NOT" include a message-body (such
	 * as the 1xx, 204, and 304 responses and any response to a HEAD
	 * request) is always terminated by the first empty line after the
	 * header fields, regardless of the entity-header fields present in
	 * the message.
	 */
	if (req->method != HTTP_M_HEAD) {
		if ((resp.status_code < 200)
		    || (resp.status_code == 204 || resp.status_code == 304)) {
			ret = 0;
			goto exit;
		}

		ret = relay_http_message_body (csock, &read_ctl_blk, &resp.mime);
		if (resp.mime.sep) {
			kfree (resp.mime.sep);
		}
	}

  exit:
	LeaveFunction(5);
	return ret;
}

int atoi(char *p)
{
   register int n;
   register int f;

   n = 0;
   f = 0;
   for(;;p++)
   {
	   switch(*p) {
		   case ' ':
		   case '\t':
		            continue;
		   case '-':
		            f++;
		   case '+':
		            p++;
		     }
		    break;
   }
   
   while(*p >= '0' && *p <= '9')
        n = n*10 + *p++ - '0';

   return(f? -n: n);
}

__u32 iptoaddr(char *sNet)
{
   int byteValue = 0;
   unsigned int subnet = atoi(sNet);
   int ret = 1;

   if (subnet > 255)
      {
       printk("Error IpAddress Bytes  %d cannot be > 255\n", subnet);
       ret = -1;
       goto out;
      }

   while (*sNet)
       {
         if (*sNet == '.')
	    {
	      byteValue = atoi(++sNet);

              if (byteValue > 255)
                 {
                   printk("Error IPAdress Bytes %d cannot be > 255\n", byteValue);
                   ret = -1;
                   goto out;
                 }

	      subnet = subnet << 8;
	      subnet = subnet + byteValue;
	    }
	    else
	        ++sNet;
	}

  out:
     if (ret == -1)
         return 0;

     return (subnet);
}

//#define NIPQUAD(addr) (int)(((addr) >> 24) & 0xff), (int)(((addr) >> 16) & 0xff), (int)(((addr) >> 8) & 0xff), (int)(((addr) >> 0) & 0xff)




static int 
tcp_vs_wlc_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
{
	register struct list_head *l, *e;
	struct tcp_vs_dest *dest, *least;

	TCP_VS_DBG(5, "tcp_vs_wlc_schedule(): Scheduling...\n");

	/*
	 * We use the following formula to estimate the overhead:
	 *                dest->conns / dest->weight
	 *
	 * Remember -- no floats in kernel mode!!!
	 * The comparison of h1*w2 > h2*w1 is equivalent to that of
	 *                h1/w1 > h2/w2
	 * when each weight is larger than zero.
	 *
	 * The server with weight=0 is quiesced and will not receive any
	 * new connection.
	 */

	read_lock(&svc->lock);
	l = &svc->destinations;
	for (e=l->next; e!=l; e=e->next) {
		least = list_entry(e, struct tcp_vs_dest, n_list);
		if (least->weight > 0) {
			goto nextstage;
		}
	}
	read_unlock(&svc->lock);
	return -1;

	/*
	 *    Find the destination with the least load.
	 */
  nextstage:
	for (e=e->next; e!=l; e=e->next) {
		dest = list_entry(e, struct tcp_vs_dest, n_list);
		if (atomic_read(&least->conns)*dest->weight
		    > atomic_read(&dest->conns)*least->weight) {
			least = dest;
		}
	}
	read_unlock(&svc->lock);

	TCP_VS_DBG(5, "WLC: server %d.%d.%d.%d:%d "
		   "conns %d refcnt %d weight %d\n",
		   NIPQUAD(least->addr), ntohs(least->port),
		   atomic_read(&least->conns),
		   atomic_read(&least->refcnt), least->weight);

	conn->dsock = tcp_vs_connect2dest(least);
	if (!conn->dsock) {
		TCP_VS_ERR_RL("The destination is not available\n");
		return -1;
	}
	atomic_inc(&least->conns);
	conn->dest = least;

	return 0;
}





static int tcp_vs_lcsphttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
{
	http_request_t		req;
	http_read_ctl_block_t	read_ctl_blk;
	char*			buffer = NULL;	/* store data from server */
        char*			header = NULL;   /* store header data from client */
	char*			body = NULL;
	int 			header_length=0;
	int			ret = 1;	/* scheduler has done all the jobs */
	int			len;
	unsigned long		last_read;
	int			close_server = 0;
	struct tcp_vs_dest	*dest;
	struct tcp_vs_dest	xdest;
	struct socket		*dsock;
//	server_conn_t		*sc;
	char*			body_decrypt;

	/* DECLARE_WAIT_QUEUE_HEAD(WQ); */
	DECLARE_WAITQUEUE(wait, current);

	EnterFunction(5);

	memset (&read_ctl_blk, 0, sizeof (read_ctl_blk));
	read_ctl_blk.buffer	= conn->buffer;
	read_ctl_blk.len	= conn->buflen;
	read_ctl_blk.flag	= MSG_MORE;
	read_ctl_blk.sock	= conn->csock;

	conn->dest	= NULL;
	conn->dsock	= NULL;

	if (read_ctl_blk.len < HTTP_MAX_REQLINE_SIZE) {
		TCP_VS_ERR("Error: buffer is too small !read_ctl_blk.len ::%d  HTTP_MAX_REQLINE_SIZE ::%d\n", read_ctl_blk.len, HTTP_MAX_REQLINE_SIZE); 
		goto out;
	}

	/* Do we have data ? */
	while (skb_queue_empty(&(conn->csock->sk->receive_queue))) {
		interruptible_sleep_on_timeout(&conn->csock->wait, HZ);
	}

	/* allocate buffer to store data that get from servers */
	buffer = (char*) get_free_page(GFP_KERNEL);
	if (buffer == NULL) {
		goto out;
	}

	header = buffer;
		
	if ( HTTP_MAX_HEADER_SIZE > PAGE_SIZE)
	{
		if (!(header = kmalloc(HTTP_MAX_HEADER_SIZE, GFP_ATOMIC))) 
		{
			TCP_VS_ERR("kmalloc failed.\n");
			ret = -2;
			goto out;
		}
	}

	last_read = jiffies;


//	do {
		switch (data_available(&read_ctl_blk)) {
		case -1:
			TCP_VS_DBG(5, "Socket error before reading request line.\n");
			ret = -2;
			goto out;

		case 0:
			/* check if the service is stopped or system is
			   unloaded */
			if (svc->stop != 0 || sysctl_ktcpvs_unload != 0) {
				TCP_VS_DBG(5, "lcsphttp scheduling exit (pid=%d)\n",
					   current->pid);
				goto out;
			}

			if ((jiffies - last_read) >
			    sysctl_ktcpvs_keepalive_timeout * HZ) {
				TCP_VS_DBG(5, "Timeout, disconnect.\n");
				goto out;
			}

			/* interruptible_sleep_on_timeout(&WQ, HZ); */
			add_wait_queue(read_ctl_blk.sock->sk->sleep, &wait);
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5*HZ);
			__set_current_state(TASK_RUNNING);
			remove_wait_queue(read_ctl_blk.sock->sk->sleep, &wait);
		//	continue;

		case 1:
			last_read = jiffies;
			break;
		}

		/* read request line from client socket */
		len = http_read_line(&read_ctl_blk, HTTP_MAX_REQLINE_SIZE);
		if (len < 0) {
			TCP_VS_ERR("Error reading request line from client\n");
			ret = -2;
			goto out;
		}

		/* parse the http request line */
		memset(&req, 0, sizeof (req));
		if (parse_http_request_line(read_ctl_blk.info, len, &req) != PARSE_OK) {
			TCP_VS_ERR("Cannot parse http request\n");
			ret = -2;
			goto out;
		}

		header_length = len + 2;
		strncpy(header, read_ctl_blk.info, len+2);

		/* read in the mime data */ 
		do {
			len = http_read_line (&read_ctl_blk, HTTP_MAX_MIMLINE_SIZE);
			if (len < 0) {
				goto out;
			}

			header_length += (len + 2);
			if (header_length >= HTTP_MAX_HEADER_SIZE)
			{
				goto out;
			}
			strncat(header, read_ctl_blk.info, len+2);
			http_line_unescape (read_ctl_blk.info, len);
			http_mime_parse (read_ctl_blk.info, len, &req.mime, 0);
		} while (len != 0); /* http header end with CRLF,CRLF */


		printk("headers %s", header);

		if (req.method == HTTP_M_POST) {
			if (req.mime.content_length > 0){
				// allocate memory
				body = header+header_length;
				
				if (req.mime.content_length + header_length > PAGE_SIZE)
				{
					if (!(body = kmalloc(req.mime.content_length, GFP_ATOMIC))) {
						TCP_VS_ERR("kmalloc failed.\n");
						ret = -2;
						goto out;
					}
				}
		}

			memset(body,'\0',req.mime.content_length);

			//retrive the data
			if (get_http_body(body, &read_ctl_blk, req.mime.content_length) != 0) {
				TCP_VS_ERR("Error in sending http message body\n");
				goto out_free;
			}
			body[req.mime.content_length]='\0';
			
		}



		//perform the rulematchine 
		/* select a server */
		dest =  tcp_vs_phttp_matchrule(svc, &req, conn, NULL, body);
		if (!dest) {
			TCP_VS_DBG(5, "Can't find a right server\n");
//			if (read_ctl_blk.flag == MSG_PEEK) {
//				/* redirect to a local port */
//				ret = -1;
//			} else {
				ret = -2;
//			}
			goto out;
		}


		printk("addr %x port %d\n",dest->addr, dest->port);
		printk("client addr %x.%d\n", conn->addr,conn->port);
#if 0

		//route request to the selected server
		if (req.version <= HTTP_V_1_0) {
			TCP_VS_DBG(5, "HTTP: server %d.%d.%d.%d:%d "
				   "conns %d refcnt %d weight %d\n",
				   NIPQUAD(dest->addr), ntohs(dest->port),
				   atomic_read(&dest->conns),
				   atomic_read (&dest->refcnt), dest->weight);

			conn->dsock = tcp_vs_connect2dest(dest);
			if (conn->dsock == NULL) {
				TCP_VS_ERR("The destination is not available!\n");
				goto out;
			}

			atomic_inc(&dest->conns);
			conn->dest = dest;
			ret = 0;
			goto out;
		}
		/*
		 * For http 1.1 client, continue processing for
		 * persistent connections
		 */

		/* lookup a server connection in the connection pool */
	  lookup_again:
		sc = tcp_vs_srvconn_get(dest->addr, dest->port);
		TCP_VS_DBG(5," sc == NULL is %s \n",(sc == NULL) ?"true" : "false");

		if (sc == NULL) {
			sc = tcp_vs_srvconn_new(dest);
			if (sc == NULL) {
//				if (read_ctl_blk.flag == MSG_PEEK) {
//					/* redirect to a local port */
//					ret = -1;
//				} else {
					ret = -2;
//				}
				goto out;
			}
		}
		dsock = sc->sock;
		if (dsock->sk->state != TCP_ESTABLISHED) {
			tcp_vs_srvconn_free(sc);
			goto lookup_again;
		}

#else
		dsock = tcp_vs_connect2dest(dest);
			if (!dsock) {
				TCP_VS_ERR_RL("The destination is not available\n");
				return -1;
			}

		atomic_inc(&dest->conns);
		conn->dest = dest;
		conn->dsock = dsock;


#endif

		if (tcp_vs_sendbuffer(dsock, header, header_length, 0) != header_length) {
			TCP_VS_ERR_RL("Error HTTP sending buffer\n");
		}

		printk("if (req.mime.content_length(%d) > 0)\n",req.mime.content_length) ;

		if (req.mime.content_length > 0) {
			if (tcp_vs_sendbuffer(dsock, body, req.mime.content_length, 0) != req.mime.content_length) {
				TCP_VS_ERR_RL("Error HTTP sending buffer\n");
			}
		}
		
#if 0
		if (tcp_vs_sendbuffer(dsock, body, header_length, 0) != header_length) {
			TCP_VS_ERR_RL("Error HTTP sending buffer\n");
		}

		// send data back to the server
		if (forward_http_message(dsock, header,header_length, body, &req.mime) != 0) {
			TCP_VS_ERR("Error in sending http message body\n");
			goto out_free;
		}
#endif


		if (http_get_response(conn->csock, dsock, &req, buffer,
				      PAGE_SIZE, &close_server) < 0) {
			goto out_free;
		}


		if (close_server) {
			TCP_VS_DBG(5, "Close server connection.\n");
			goto out_free;	/* close the connection? tbd */
		}

		/* have to put back the server connection when
		 * 		   it is not used */
//		tcp_vs_srvconn_put(sc);
		
//	} while (req.mime.connection_close != 1);

 out:
	if (buffer) {
		free_page ((unsigned long)buffer);
	}

	if (req.mime.content_length + header_length > PAGE_SIZE) {
		kfree(body);
	}
	
	if ( HTTP_MAX_HEADER_SIZE > PAGE_SIZE)
	{
		kfree(header);
	}
	
	LeaveFunction(5);
	return ret;

 out_free:
//	tcp_vs_srvconn_free(sc);
	goto out;
}


static struct tcp_vs_scheduler tcp_vs_lcsphttp_scheduler =
{
	{0},				/* n_list */
	"lcsphttp",			/* name */
	THIS_MODULE,			/* this module */
	tcp_vs_lcsphttp_init_svc,	/* initializer */
	tcp_vs_lcsphttp_done_svc,	/* done */
	tcp_vs_lcsphttp_update_svc,	/* update */
	tcp_vs_lcsphttp_schedule,	/* select a server by http request */
};


static int __init tcp_vs_lcsphttp_init(void)
{
	INIT_LIST_HEAD(&tcp_vs_lcsphttp_scheduler.n_list);
	return register_tcp_vs_scheduler(&tcp_vs_lcsphttp_scheduler);
}

static void __exit tcp_vs_lcsphttp_cleanup(void)
{
	unregister_tcp_vs_scheduler(&tcp_vs_lcsphttp_scheduler);
}

module_init(tcp_vs_lcsphttp_init);
module_exit(tcp_vs_lcsphttp_cleanup);
MODULE_LICENSE("GPL");
